# coding: utf-8

# psim を利用するためのおまじない -- ここから
import winreg, sys, os
reg = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Wow6432Node\\Mathematical Systems Inc.\\S-Quattro Simulation System\\INFO")
rootdir = winreg.QueryValueEx(reg, "ROOTDIR")[0]
os.environ["PATH"] = rootdir + ";" + os.environ["PATH"]
sys.path.insert(0, os.path.join(rootdir, "lib3"))
# psim を利用するためのおまじない -- ここまで

import matplotlib.pyplot as plt
import xlwings as xw

from psim import *

class Graph:
    def __init__(self, sheet, plotname, cell, width = 360, height = 360):
        self.sheet = sheet
        self.plotname = plotname
        self.cell = cell
        self.width = width
        self.height = height
    def __enter__(self):
        dpi = plt.rcParams["figure.dpi"]
        if self.plotname in self.sheet.pictures:
            pic = self.sheet.pictures[self.plotname]
            figsize = (pic.width / dpi, pic.height / dpi)
        else:
            figsize = (self.width / dpi, self.height / dpi)
        fig, ax = plt.subplots(1, 1, figsize = figsize)
        return fig, ax
    def __exit__(self, exc_type, exc_value, traceback):
        dirname = os.path.dirname(self.sheet.book.fullname)
        pngname = os.path.join(dirname, "temp.png")
        plt.tight_layout()
        plt.savefig(pngname)
        self.sheet.pictures.add(pngname,
                                name = self.plotname,
                                update = True,
                                left = self.sheet.range(self.cell).left,
                                top = self.sheet.range(self.cell).top)
    
def runSimulation():
    book = xw.Book.caller()
    sheet = book.sheets[0]
    plt.rcParams['font.sans-serif'].insert(0, "Meiryo")

    initialize()
    
    lmd = xw.Range("B3").value # 平均到着率
    mu = xw.Range("B4").value # 平均サービス率
    simtime = xw.Range("B5").value # シミュレーション時間

    # 到着間隔を生成する乱数発生器
    g1 = exponentialDistribution(1 / lmd)
    # サービス時間を生成する乱数発生器
    g2 = exponentialDistribution(1 / mu)
    # 窓口を示すファシリティ資源
    f = Facility(1, monitor = True)
    # 待ち時間を記録するモニター
    monitor = Monitor(["waitingTime"], ["f"], name = "wt")

    # 到着プロセス
    def arrive():
        while True:
            # 次の到着まで待つ
            yield pause(next(g1))
            # サービスプロセスを開始する
            # - この処理自体は非同期なので、一瞬で完了する
            yield subactivate(service)()

    # サービスプロセス
    def service():
        t0 = now()
        # 窓口の待ち行列に並ぶ
        # - 窓口を示すファシリティ資源の利用を予約する
        # - 既に利用者がいる場合は、利用可能になるまで待ち受ける
        lock = (yield f.request(name = "lock"))["lock"]
        # 窓口サービスを開始する
        # - 窓口を示すファシリティ資源の利用権を取得した
        # - 利用権を開放するまで他の利用者は利用できない
        t1 = now()
        # 待ち時間を記録
        monitor.observe(t1 - t0)
        # 窓口を利用する
        yield pause(next(g2))
        # 窓口を示すファシリティ資源の利用権を開放する
        lock.release()

    # 到着プロセスの開始を登録する
    activate(arrive)()

    # シミュレーション開始
    start(until = simtime)

    # M/M/1 の理論値
    rho = lmd / mu # 平均利用率
    Q = rho / (1.0 - rho) # 平均待ち行列
    W = Q / mu # 平均待ち時間

    waitingTime = monitor["waitingTime"].mean()
    # 平均待ち行列(窓口利用者をカウントしない)
    queue0 = f.monitor["要求待ち"].mean()
    # 平均システム内客数(窓口利用者をカウントする)
    queue = (f.monitor["要求待ち"] + f.monitor["バッファ"]).mean()
    #print(f"平均待ち時間: {waitingTime:.3f} (理論値: {W:.3f})")
    #print(f"平均システム内客数: {queue:.3f} (理論値: {Q:.3f})")
    xw.Range("B9").value = waitingTime
    xw.Range("B10").value = queue
    xw.Range("D9").value = W
    xw.Range("D10").value = Q

    with Graph(sheet, "Plot1", "G2", width = 720, height = 360) as (fig, ax):
        x = f.monitor.time().toList()
        y = (f.monitor["要求待ち"] + f.monitor["バッファ"]).toList()
        ax.plot(x, y, label = "システム内客数")
        ax.set_xlabel("時間")
        ax.set_ylabel("システム内客数")
        ax.set_title("システム内客数")
        ax.legend()

    with Graph(sheet, "Plot2", "G17") as (fig, ax):
        ax.hist(monitor["waitingTime"].toList())
        ax.set_xlabel("待ち時間(秒)")
        ax.set_ylabel("頻度")
        ax.set_title("待ち時間の分布")
    
    with Graph(sheet, "Plot3", "L17") as (fig, ax):
        w = f.monitor.time().diff()
        v = (f.monitor["要求待ち"] + f.monitor["バッファ"]).toList()[:-1]
        ax.hist(v, weights = w, density = True)
        ax.set_xlabel("システム内客数")
        ax.set_ylabel("密度")
        ax.set_title("システム内客数の分布")
    
def clear():
    book = xw.Book.caller()
    sheet = book.sheets[0]
    xw.Range("B9").value = ""
    xw.Range("B10").value = ""
    xw.Range("D9").value = ""
    xw.Range("D10").value = ""
    for plotname in ["Plot1", "Plot2", "Plot3"]:
        if plotname in sheet.pictures:
            pic = sheet.pictures[plotname]
            pic.delete()
